#ifndef MEASURE_H
#define MEASURE_H

#ifndef RDTSC
#define MEASURE(x) (x)
#endif

/* This part defines the functions and MACROS needed to measure using RDTSC */
#ifdef RDTSC
   
   #ifndef REPEAT     
      #define REPEAT 10000
   #endif
   
   #ifndef OUTER_REPEAT
      #define OUTER_REPEAT 30
   #endif

   #ifndef WARMUP
      #define WARMUP REPEAT/4
   #endif

   unsigned long long RDTSC_start_clk, RDTSC_end_clk;
   double RDTSC_total_clk;
	double RDTSC_TEMP_CLK;
   int RDTSC_MEASURE_ITERATOR;
	int RDTSC_OUTER_ITERATOR;

   inline static unsigned long get_Clks(void)
   {
      unsigned long long ret_val;
      __asm__ volatile
         (
         "cpuid\n\t\
          rdtsc\n\t\
          mov %%eax,(%0)\n\t\
          mov %%edx,4(%0)"::"rm"(&ret_val):"eax","edx","ebx","ecx"
          );
	   return ret_val;
   }

   /* 
   This MACRO measures the number of cycles "x" runs. This is the flow:
      1) it sets the priority to FIFO, to avoid time slicing if possible.
      2) it repeats "x" WARMUP times, in order to warm the cache.
      3) it reads the Time Stamp Counter at the beginning of the test.
      4) it repeats "x" REPEAT number of times.
      5) it reads the Time Stamp Counter again at the end of the test
      6) it calculates the average number of cycles per one iteration of "x", by calculating the total number of cycles, and dividing it by REPEAT
    */      
   #define RDTSC_MEASURE(x)                                                                        \
   for(RDTSC_MEASURE_ITERATOR=0; RDTSC_MEASURE_ITERATOR< WARMUP; RDTSC_MEASURE_ITERATOR++)         \
      {                                                                                            \
         {x};                                                                                      \
      }                                                                                    		   \
	RDTSC_total_clk = 1.7976931348623157e+308;																	   \
	for(RDTSC_OUTER_ITERATOR=0;RDTSC_OUTER_ITERATOR<OUTER_REPEAT; RDTSC_OUTER_ITERATOR++){			   \
      RDTSC_start_clk = get_Clks();                                                                \
      for (RDTSC_MEASURE_ITERATOR = 0; RDTSC_MEASURE_ITERATOR < REPEAT; RDTSC_MEASURE_ITERATOR++)  \
      {                                                                                            \
         {x};                                                                                      \
      }                                                                                            \
      RDTSC_end_clk = get_Clks();   				                                               		\
      RDTSC_TEMP_CLK = (double)(RDTSC_end_clk-RDTSC_start_clk)/REPEAT;										\
		if(RDTSC_total_clk>RDTSC_TEMP_CLK) RDTSC_total_clk = RDTSC_TEMP_CLK;									\
	}

	#define MEASURE(x) RDTSC_MEASURE(x)
	
#endif

#endif